/* SPDX-License-Identifier: GPL-2.0 */
/*
 * Copyright (C) 2013 Imagination Technologies
 * Author: Paul Burton <paul.burton@mips.com>
 * Copyright (C) 2020 Loongson Technology Co., Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */
#ifndef _ASM_LSX_H
#define _ASM_LSX_H

#include <asm/loongarchregs.h>

#ifndef __ASSEMBLY__

#include <asm/inst.h>
#include <lcsrintrin.h>

extern void _save_lsx(struct loongarch_fpu *fpu);
extern void _restore_lsx(struct loongarch_fpu *fpu);
extern void _init_lsx_upper(void);
extern void _restore_lsx_upper(struct loongarch_fpu *fpu);

#ifdef CONFIG_LOONGSON3_ACPI_CPUFREQ
DECLARE_PER_CPU(unsigned long, msa_count);
#endif

/*
 * SIMD 2 bit format fields.
 */
enum simd_2b_fmt {
	simd_fmt_b = 0,
	simd_fmt_h = 1,
	simd_fmt_w = 2,
	simd_fmt_d = 3,
};

#ifdef CONFIG_CPU_HAS_LASX
enum {
	CTX_LSX = 1,
	CTX_LASX = 2,
};

extern void _save_lasx(struct loongarch_fpu *fpu);
extern void _restore_lasx(struct loongarch_fpu *fpu);
extern void _init_lasx_upper(void);
extern void _restore_lasx_upper(struct loongarch_fpu *fpu);

#ifdef CONFIG_LOONGSON3_ACPI_CPUFREQ
DECLARE_PER_CPU(unsigned long, lasx_count);
#endif
static inline int is_lasx_enabled(void)
{
	if (!cpu_has_lasx)
		return 0;

	return (csr_readl(LOONGARCH_CSR_EUEN) & CSR_EUEN_LASXEN) ?
			1 : 0;
}

static inline void enable_lasx(void)
{

	if (cpu_has_lasx) {
		csr_xchgl(CSR_EUEN_LASXEN, CSR_EUEN_LASXEN, LOONGARCH_CSR_EUEN);
#ifdef CONFIG_LOONGSON3_ACPI_CPUFREQ
		per_cpu(lasx_count, raw_smp_processor_id())++;
#endif
	}
}

static inline void disable_lasx(void)
{
	if (cpu_has_lasx) {
		csr_xchgl(0, CSR_EUEN_LASXEN, LOONGARCH_CSR_EUEN);
	}
}

static inline void save_lasx(struct task_struct *t)
{
	if (cpu_has_lasx)
		_save_lasx(&t->thread.fpu);
}

static inline void restore_lasx(struct task_struct *t)
{
	if (cpu_has_lasx)
		_restore_lasx(&t->thread.fpu);
}

static inline void init_lasx_upper(void)
{
	if (cpu_has_lasx)
		_init_lasx_upper();
}

static inline void restore_lasx_upper(struct task_struct *t)
{
	if (cpu_has_lasx)
		_restore_lasx_upper(&t->thread.fpu);
}
#endif

static inline int is_lsx_owner(void)
{
	return test_thread_flag(TIF_USEDSIMD);
}

extern void read_lsx_wr_b(unsigned idx, union fpureg *to);
extern void read_lsx_wr_h(unsigned idx, union fpureg *to);
extern void read_lsx_wr_w(unsigned idx, union fpureg *to);
extern void read_lsx_wr_d(unsigned idx, union fpureg *to);

/**
 * read_lsx_wr() - Read a single LSX vector register
 * @idx:	The index of the vector register to read
 * @to:		The FPU register union to store the registers value in
 * @fmt:	The format of the data in the vector register
 *
 * Read the value of LSX vector register idx into the FPU register
 * union to, using the format fmt.
 */
static inline void read_lsx_wr(unsigned idx, union fpureg *to,
			       enum simd_2b_fmt fmt)
{
	switch (fmt) {
	case simd_fmt_b:
		read_lsx_wr_b(idx, to);
		break;

	case simd_fmt_h:
		read_lsx_wr_h(idx, to);
		break;

	case simd_fmt_w:
		read_lsx_wr_w(idx, to);
		break;

	case simd_fmt_d:
		read_lsx_wr_d(idx, to);
		break;

	default:
		BUG();
	}
}

extern void write_lsx_wr_b(unsigned idx, union fpureg *from);
extern void write_lsx_wr_h(unsigned idx, union fpureg *from);
extern void write_lsx_wr_w(unsigned idx, union fpureg *from);
extern void write_lsx_wr_d(unsigned idx, union fpureg *from);

/**
 * write_lsx_wr() - Write a single LSX vector register
 * @idx:	The index of the vector register to write
 * @from:	The FPU register union to take the registers value from
 * @fmt:	The format of the data in the vector register
 *
 * Write the value from the FPU register union from into LSX vector
 * register idx, using the format fmt.
 */
static inline void write_lsx_wr(unsigned idx, union fpureg *from,
				enum simd_2b_fmt fmt)
{
	switch (fmt) {
	case simd_fmt_b:
		write_lsx_wr_b(idx, from);
		break;

	case simd_fmt_h:
		write_lsx_wr_h(idx, from);
		break;

	case simd_fmt_w:
		write_lsx_wr_w(idx, from);
		break;

	case simd_fmt_d:
		write_lsx_wr_d(idx, from);
		break;

	default:
		BUG();
	}
}

static inline void enable_lsx(void)
{
	if (cpu_has_lsx) {
		csr_xchgl(CSR_EUEN_LSXEN, CSR_EUEN_LSXEN, LOONGARCH_CSR_EUEN);
#ifdef CONFIG_LOONGSON3_ACPI_CPUFREQ
		per_cpu(msa_count, raw_smp_processor_id())++;
#endif
	}
}

static inline void disable_lsx(void)
{
	if (cpu_has_lsx) {
		csr_xchgl(0, CSR_EUEN_LSXEN, LOONGARCH_CSR_EUEN);
	}
}

static inline int is_lsx_enabled(void)
{
	if (!cpu_has_lsx)
		return 0;

	return (csr_readl(LOONGARCH_CSR_EUEN) & CSR_EUEN_LSXEN) ?
		1 : 0;
}

static inline int is_simd_enabled(void)
{
	return is_lsx_enabled() | is_lasx_enabled();
}

static inline int is_fp_enabled(void)
{
	return (csr_readl(LOONGARCH_CSR_EUEN) & CSR_EUEN_FPEN) ?
		1 : 0;
}

static inline int thread_simd_context_live(void)
{
	int ret = 0;
	/*
	 * Check cpu_has_lsx only if it's a constant. This will allow the
	 * compiler to optimise out code for CPUs without LSX without adding
	 * an extra redundant check for CPUs with LSX.
	 */
	if (__builtin_constant_p(cpu_has_lsx) && !cpu_has_lsx)
#ifndef CONFIG_CPU_HAS_LASX
		return 0;
	return test_thread_flag(TIF_LSX_CTX_LIVE);
#else
		goto  out;

	ret =  test_thread_flag(TIF_LSX_CTX_LIVE) ? CTX_LSX : 0;

	if (__builtin_constant_p(cpu_has_lasx) && !cpu_has_lasx)
		goto out;

	ret = test_thread_flag(TIF_LASX_CTX_LIVE) ? CTX_LASX : ret;
out:
	return ret;
#endif
}

static inline int thread_lsx_context_live(void)
{
	int ret = 0;
	/*
	 * Check cpu_has_lsx only if it's a constant. This will allow the
	 * compiler to optimise out code for CPUs without LSX without adding
	 * an extra redundant check for CPUs with LSX.
	 */
	if (__builtin_constant_p(cpu_has_lsx) && !cpu_has_lsx)
		goto  out;

	ret =  test_thread_flag(TIF_LSX_CTX_LIVE) ? CTX_LSX : 0;
out:
	return ret;
}

static inline int thread_lasx_context_live(void)
{
	int ret = 0;

	if (__builtin_constant_p(cpu_has_lasx)
		&& !cpu_has_lasx)
		goto out;

	ret = test_thread_flag(TIF_LASX_CTX_LIVE) ? CTX_LASX : ret;
out:
	return ret;
}

static inline void save_lsx(struct task_struct *t)
{
	if (cpu_has_lsx)
		_save_lsx(&t->thread.fpu);
}

static inline void restore_lsx(struct task_struct *t)
{
	if (cpu_has_lsx)
		_restore_lsx(&t->thread.fpu);
}

static inline void restore_lsx_upper(struct task_struct *t)
{
	if (cpu_has_lsx)
		_restore_lsx_upper(&t->thread.fpu);
}

static inline void init_lsx_upper(void)
{
	/*
	 * Check cpu_has_lsx only if it's a constant. This will allow the
	 * compiler to optimise out code for CPUs without LSX without adding
	 * an extra redundant check for CPUs with LSX.
	 */
	if (__builtin_constant_p(cpu_has_lsx) && !cpu_has_lsx)
		return;

	_init_lsx_upper();
}

#define __BUILD_LSX_CTL_REG(name, cs)				\
static inline unsigned int read_v##name(void)		\
{								\
	unsigned int reg;					\
	__asm__ __volatile__(					\
	"	movfcsr2gr	%0, $r" #cs "\n"			\
	: "=r"(reg));						\
	return reg;						\
}								\
								\
static inline void write_v##name(unsigned int val)		\
{								\
	__asm__ __volatile__(					\
	"	movgr2fcsr	$r" #cs ", %0\n"			\
	: : "r"(val));						\
}

__BUILD_LSX_CTL_REG(csr0, 0)
__BUILD_LSX_CTL_REG(csr1, 1)
__BUILD_LSX_CTL_REG(csr2, 2)
__BUILD_LSX_CTL_REG(csr3, 3)
__BUILD_LSX_CTL_REG(csr16, 16)

#endif /* !__ASSEMBLY__ */

#endif /* _ASM_LSX_H */
